import MultipartParser, mimetools, logging, tempfile, os, sys, shutil
import tempfile_compat
import osutil
from ReadLineAdapter import ReadLineAdapter

_logger = logging.getLogger("streaming_extractor")


class PartNotFound:
    def __init__(self, part_id):
        self.part_id = part_id


class NotAMimeError(Exception):
    def __str__(self):
        return "Supplied file is not in proper format"


class StreamingExtractor(object):
    """MIME-file parts extractor.

    Given the filehandle, reads the MIME-message from this handle,
    unpacking the requested parts. If requests for extraction will
    arrive in the order different from the order of parts, read parts
    will be put to the temporary files until request.
    """

    def __init__(self, fp, session_dir, dump_file = None):
        """Constructs the extractor from the given filehandle"""

        tempfile.tempdir = session_dir

        msg = mimetools.Message(fp)
        if msg.gettype() != 'multipart/related':
            raise NotAMimeError()
        boundary = msg.getparam("boundary")

        try:
            self.mf = MultipartParser.MultipartParser(fp, boundary)
        except Exception:
            raise NotAMimeError()

        self.saved_parts = {}

        amf = ReadLineAdapter(self.mf)
        submsg = mimetools.Message(amf)

        if dump_file:
            encoding = submsg.getheader("Content-Transfer-Encoding")
            if encoding == "binary":
                encoding = "8bit"
            df = file(dump_file, "w")
            try:
                try:
                    mimetools.decode(amf, df, encoding)
                except Exception:
                    raise NotAMimeError()
            finally:
                df.close()

    def _skip_until(self, id):
        """Reads the file until MIME-part with Content-Id: <id>

        Raises PartNotFound(id) if part is not found during search.
        Skipped parts are saved to the temporary directories.
        """

        _logger.info("Searching for MIME-part %s" % id)

        while self.mf.next():
            amf = ReadLineAdapter(self.mf)
            submsg = mimetools.Message(amf)
            part_id = submsg.getheader("content-id")
            if part_id == id:
                return (submsg, amf)
            else:
                self._save_part(submsg, amf)
        raise PartNotFound(id)

    def _save_part(self, submsg, mf):
        """Saves current part of MIME-file to the file.

        Registers the file in the internal structures for later
        retrieval by extract_saved_part
        """
        id = submsg.getheader("Content-Id")
        encoding = submsg.getheader("Content-Transfer-Encoding")

        part_file_name = tempfile.mktemp()

        _logger.info("MIME-part %s -> %s" % (id, part_file_name))

        part_file = file(part_file_name, "w")
        if encoding == "binary" or encoding == "8bit":
            shutil.copyfileobj(mf, part_file)
        else:
            mimetools.decode(mf, part_file, encoding)
        part_file.close()

        self.saved_parts[id] = (part_file_name, encoding)

    def _extract_saved_part(self, id, unpacker):
        """Extracts the saved to the disk part 'id' to the unpacker."""

        try:
            (filename, encoding) = self.saved_parts[id]

            _logger.info("Processing MIME-part %s (%s)" % (id, filename))

            fd = file(filename)
            try:
                res = unpacker.unpack(fd, id)
            finally:
                fd.close()
        finally:
            osutil.unlink_nothrow(filename)
            del self.saved_parts[id]
        return res

    def extract_part(self, id, unpacker):
        """Extracts the MIME-part 'id'.

        The extracted part is passed to the unpacker.
        """
        if self.saved_parts.has_key(id):
            return self._extract_saved_part(id, unpacker)
        else:
            (submsg, mf) = self._skip_until(id)
            self._save_part(submsg, mf)
            return self._extract_saved_part(id, unpacker)
